@@ -1,5 +1,9 @@
 require 'net/http'
-require 'net/https'
+begin
+  require 'net/https'
+rescue LoadError
+  # net/https or openssl
+end if RUBY_VERSION < '1.9' # but only for 1.8
 require 'net/http/faster'
 require 'uri'
 require 'cgi' # for escaping
@@ -9,6 +13,8 @@ begin
 rescue LoadError
 end
 
+autoload :OpenSSL, 'openssl'
+
 ##
 # Persistent connections for Net::HTTP
 #
@@ -37,6 +43,11 @@ end
 #   # perform a GET
 #   response = http.request uri
 #
+#   # or
+#
+#   get = Net::HTTP::Get.new uri.request_uri
+#   response = http.request get
+#
 #   # create a POST
 #   post_uri = uri + 'create'
 #   post = Net::HTTP::Post.new post_uri.path
@@ -45,6 +56,10 @@ end
 #   # perform the POST, the URI is always required
 #   response http.request post_uri, post
 #
+# Note that for GET, HEAD and other requests that do not have a body you want
+# to use URI#request_uri not URI#path.  The request_uri contains the query
+# params which are sent in the body for other requests.
+#
 # == SSL
 #
 # SSL connections are automatically created depending upon the scheme of the
@@ -105,6 +120,13 @@ end
 # The amount of time allowed between reading two chunks from the socket.  Set
 # through #read_timeout
 #
+# === Max Requests
+#
+# The number of requests that should be made before opening a new connection.
+# Typically many keep-alive capable servers tune this to 100 or less, so the
+# 101st request will fail with ECONNRESET. If unset (default), this value has no
+# effect, if set, connections will be reset on the request after max_requests.
+#
 # === Open Timeout
 #
 # The amount of time to wait for a connection to be opened.  Set through
@@ -173,10 +195,30 @@ class Net::HTTP::Persistent
 
   EPOCH = Time.at 0 # :nodoc:
 
+  ##
+  # Is OpenSSL available?  This test works with autoload
+
+  HAVE_OPENSSL = defined? OpenSSL::SSL # :nodoc:
+
   ##
   # The version of Net::HTTP::Persistent you are using
 
-  VERSION = '2.7'
+  VERSION = '2.9.4'
+
+  ##
+  # Exceptions rescued for automatic retry on ruby 2.0.0.  This overlaps with
+  # the exception list for ruby 1.x.
+
+  RETRIED_EXCEPTIONS = [ # :nodoc:
+    (Net::ReadTimeout if Net.const_defined? :ReadTimeout),
+    IOError,
+    EOFError,
+    Errno::ECONNRESET,
+    Errno::ECONNABORTED,
+    Errno::EPIPE,
+    (OpenSSL::SSL::SSLError if HAVE_OPENSSL),
+    Timeout::Error,
+  ].compact
 
   ##
   # Error class for errors raised by Net::HTTP::Persistent.  Various
@@ -185,11 +227,63 @@ class Net::HTTP::Persistent
 
   class Error < StandardError; end
 
+  ##
+  # Use this method to detect the idle timeout of the host at +uri+.  The
+  # value returned can be used to configure #idle_timeout.  +max+ controls the
+  # maximum idle timeout to detect.
+  #
+  # After
+  #
+  # Idle timeout detection is performed by creating a connection then
+  # performing a HEAD request in a loop until the connection terminates
+  # waiting one additional second per loop.
+  #
+  # NOTE:  This may not work on ruby > 1.9.
+
+  def self.detect_idle_timeout uri, max = 10
+    uri = URI uri unless URI::Generic === uri
+    uri += '/'
+
+    req = Net::HTTP::Head.new uri.request_uri
+
+    http = new 'net-http-persistent detect_idle_timeout'
+
+    connection = http.connection_for uri
+
+    sleep_time = 0
+
+    loop do
+      response = connection.request req
+
+      $stderr.puts "HEAD #{uri} => #{response.code}" if $DEBUG
+
+      unless Net::HTTPOK === response then
+        raise Error, "bad response code #{response.code} detecting idle timeout"
+      end
+
+      break if sleep_time >= max
+
+      sleep_time += 1
+
+      $stderr.puts "sleeping #{sleep_time}" if $DEBUG
+      sleep sleep_time
+    end
+  rescue
+    # ignore StandardErrors, we've probably found the idle timeout.
+  ensure
+    http.shutdown
+
+    return sleep_time unless $!
+  end
+
   ##
   # This client's OpenSSL::X509::Certificate
 
   attr_reader :certificate
 
+  # For Net::HTTP parity
+  alias cert certificate
+
   ##
   # An SSL certificate authority.  Setting this will set verify_mode to
   # VERIFY_PEER.
@@ -237,6 +331,12 @@ class Net::HTTP::Persistent
 
   attr_accessor :idle_timeout
 
+  ##
+  # Maximum number of requests on a connection before it is considered expired
+  # and automatically closed.
+
+  attr_accessor :max_requests
+
   ##
   # The value sent in the Keep-Alive header.  Defaults to 30.  Not needed for
   # HTTP/1.1 servers.
@@ -269,11 +369,19 @@ class Net::HTTP::Persistent
 
   attr_reader :private_key
 
+  # For Net::HTTP parity
+  alias key private_key
+
   ##
   # The URL through which requests will be proxied
 
   attr_reader :proxy_uri
 
+  ##
+  # List of host suffixes which will not be proxied
+
+  attr_reader :no_proxy
+
   ##
   # Seconds to wait until reading one block.  See Net::HTTP#read_timeout
 
@@ -376,6 +484,7 @@ class Net::HTTP::Persistent
 
     @debug_output     = nil
     @proxy_uri        = nil
+    @no_proxy         = []
     @headers          = {}
     @override_headers = {}
     @http_versions    = {}
@@ -383,6 +492,7 @@ class Net::HTTP::Persistent
     @open_timeout     = nil
     @read_timeout     = nil
     @idle_timeout     = 5
+    @max_requests     = nil
     @socket_options   = []
 
     @socket_options << [Socket::IPPROTO_TCP, Socket::TCP_NODELAY, 1] if
@@ -399,15 +509,22 @@ class Net::HTTP::Persistent
     @private_key        = nil
     @ssl_version        = nil
     @verify_callback    = nil
-    @verify_mode        = OpenSSL::SSL::VERIFY_PEER
+    @verify_mode        = nil
     @cert_store         = nil
 
     @generation         = 0 # incremented when proxy URI changes
     @ssl_generation     = 0 # incremented when SSL session variables change
-    @reuse_ssl_sessions = OpenSSL::SSL.const_defined? :Session
+
+    if HAVE_OPENSSL then
+      @verify_mode        = OpenSSL::SSL::VERIFY_PEER
+      @reuse_ssl_sessions = OpenSSL::SSL.const_defined? :Session
+    end
 
     @retry_change_requests = false
 
+    @ruby_1 = RUBY_VERSION < '2'
+    @retried_on_ruby_2 = !@ruby_1
+
     self.proxy = proxy if proxy
   end
 
@@ -420,6 +537,9 @@ class Net::HTTP::Persistent
     reconnect_ssl
   end
 
+  # For Net::HTTP parity
+  alias cert= certificate=
+
   ##
   # Sets the SSL certificate authority file.
 
@@ -450,6 +570,8 @@ class Net::HTTP::Persistent
     timeouts = thread[@timeout_key]
 
     (0...generation).each do |old_generation|
+      next unless thread[generation_key]
+
       conns = thread[generation_key].delete old_generation
 
       conns.each_value do |conn|
@@ -472,6 +594,9 @@ class Net::HTTP::Persistent
     use_ssl = uri.scheme.downcase == 'https'
 
     if use_ssl then
+      raise Net::HTTP::Persistent::Error, 'OpenSSL is not available' unless
+        HAVE_OPENSSL
+
       ssl_generation = @ssl_generation
 
       ssl_cleanup ssl_generation
@@ -488,7 +613,7 @@ class Net::HTTP::Persistent
     net_http_args = [uri.host, uri.port]
     connection_id = net_http_args.join ':'
 
-    if @proxy_uri then
+    if @proxy_uri and not proxy_bypass? uri.host, uri.port then
       connection_id << @proxy_connection_id
       net_http_args.concat @proxy_args
     end
@@ -503,27 +628,22 @@ class Net::HTTP::Persistent
       reset connection if expired? connection
     end
 
-    unless connection.started? then
-      connection.set_debug_output @debug_output if @debug_output
-      connection.open_timeout = @open_timeout if @open_timeout
-      connection.read_timeout = @read_timeout if @read_timeout
+    start connection unless connection.started?
 
-      connection.start
-
-      socket = connection.instance_variable_get :@socket
-
-      if socket then # for fakeweb
-        @socket_options.each do |option|
-          socket.io.setsockopt(*option)
-        end
-      end
-    end
+    connection.read_timeout = @read_timeout if @read_timeout
+    connection.keep_alive_timeout = @idle_timeout if @idle_timeout && connection.respond_to?(:keep_alive_timeout=)
 
     connection
   rescue Errno::ECONNREFUSED
-    raise Error, "connection refused: #{connection.address}:#{connection.port}"
+    address = connection.proxy_address || connection.address
+    port    = connection.proxy_port    || connection.port
+
+    raise Error, "connection refused: #{address}:#{port}"
   rescue Errno::EHOSTDOWN
-    raise Error, "host down: #{connection.address}:#{connection.port}"
+    address = connection.proxy_address || connection.address
+    port    = connection.proxy_port    || connection.port
+
+    raise Error, "host down: #{address}:#{port}"
   end
 
   ##
@@ -548,10 +668,20 @@ class Net::HTTP::Persistent
   end
 
   ##
-  # Returns true if the connection should be reset due to an idle timeout,
-  # false otherwise.
+  # URI::unescape wrapper
+
+  def unescape str
+    CGI.unescape str if str
+  end
+
+
+  ##
+  # Returns true if the connection should be reset due to an idle timeout, or
+  # maximum request count, false otherwise.
 
   def expired? connection
+    requests = Thread.current[@request_key][connection.object_id]
+    return true  if     @max_requests && requests >= @max_requests
     return false unless @idle_timeout
     return true  if     @idle_timeout.zero?
 
@@ -560,6 +690,24 @@ class Net::HTTP::Persistent
     Time.now - last_used > @idle_timeout
   end
 
+  ##
+  # Starts the Net::HTTP +connection+
+
+  def start connection
+    connection.set_debug_output @debug_output if @debug_output
+    connection.open_timeout = @open_timeout if @open_timeout
+
+    connection.start
+
+    socket = connection.instance_variable_get :@socket
+
+    if socket then # for fakeweb
+      @socket_options.each do |option|
+        socket.io.setsockopt(*option)
+      end
+    end
+  end
+
   ##
   # Finishes the Net::HTTP +connection+
 
@@ -573,8 +721,11 @@ class Net::HTTP::Persistent
   end
 
   def http_class # :nodoc:
-    if [:FakeWeb, :WebMock].any? { |klass| Object.const_defined?(klass) } or
-      not @reuse_ssl_sessions then
+    if RUBY_VERSION > '2.0' then
+      Net::HTTP
+    elsif [:Artifice, :FakeWeb, :WebMock].any? { |klass|
+             Object.const_defined?(klass)
+          } or not @reuse_ssl_sessions then
         Net::HTTP
     else
       Net::HTTP::Persistent::SSLReuse
@@ -600,10 +751,15 @@ class Net::HTTP::Persistent
   end
 
   ##
-  # Is the request idempotent or is retry_change_requests allowed
+  # Is the request +req+ idempotent or is retry_change_requests allowed.
+  #
+  # If +retried_on_ruby_2+ is true, true will be returned if we are on ruby,
+  # retry_change_requests is allowed and the request is not idempotent.
 
-  def can_retry? req
-    retry_change_requests or idempotent?(req)
+  def can_retry? req, retried_on_ruby_2 = false
+    return @retry_change_requests && !idempotent?(req) if retried_on_ruby_2
+
+    @retry_change_requests || idempotent?(req)
   end
 
   if RUBY_VERSION > '1.9' then
@@ -680,6 +836,9 @@ class Net::HTTP::Persistent
     reconnect_ssl
   end
 
+  # For Net::HTTP parity
+  alias key= private_key=
+
   ##
   # Sets the proxy server.  The +proxy+ may be the URI of the proxy server,
   # the symbol +:ENV+ which will read the proxy from the environment or nil to
@@ -689,8 +848,10 @@ class Net::HTTP::Persistent
   # If the proxy URI is set after requests have been made, the next request
   # will shut-down and re-open all connections.
   #
-  # If you are making some requests through a proxy and others without a proxy
-  # use separate Net::Http::Persistent instances.
+  # The +no_proxy+ query parameter can be used to specify hosts which shouldn't
+  # be reached via proxy; if set it should be a comma separated list of
+  # hostname suffixes, optionally with +:port+ appended, for example
+  # <tt>example.com,some.host:8080</tt>.
 
   def proxy= proxy
     @proxy_uri = case proxy
@@ -700,15 +861,21 @@ class Net::HTTP::Persistent
                  else raise ArgumentError, 'proxy must be :ENV or a URI::HTTP'
                  end
 
+    @no_proxy.clear
+
     if @proxy_uri then
       @proxy_args = [
         @proxy_uri.host,
         @proxy_uri.port,
-        @proxy_uri.user,
-        @proxy_uri.password,
+        unescape(@proxy_uri.user),
+        unescape(@proxy_uri.password),
       ]
 
       @proxy_connection_id = [nil, *@proxy_args].join ':'
+
+      if @proxy_uri.query then
+        @no_proxy = CGI.parse(@proxy_uri.query)['no_proxy'].join(',').downcase.split(',').map { |x| x.strip }.reject { |x| x.empty? }
+      end
     end
 
     reconnect
@@ -724,6 +891,12 @@ class Net::HTTP::Persistent
   # indicated user and password unless HTTP_PROXY contains either of these in
   # the URI.
   #
+  # The +NO_PROXY+ ENV variable can be used to specify hosts which shouldn't
+  # be reached via proxy; if set it should be a comma separated list of
+  # hostname suffixes, optionally with +:port+ appended, for example
+  # <tt>example.com,some.host:8080</tt>. When set to <tt>*</tt> no proxy will
+  # be returned.
+  #
   # For Windows users, lowercase ENV variables are preferred over uppercase ENV
   # variables.
 
@@ -734,6 +907,15 @@ class Net::HTTP::Persistent
 
     uri = URI normalize_uri env_proxy
 
+    env_no_proxy = ENV['no_proxy'] || ENV['NO_PROXY']
+
+    # '*' is special case for always bypass
+    return nil if env_no_proxy == '*'
+
+    if env_no_proxy then
+      uri.query = "no_proxy=#{escape(env_no_proxy)}"
+    end
+
     unless uri.user or uri.password then
       uri.user     = escape ENV['http_proxy_user'] || ENV['HTTP_PROXY_USER']
       uri.password = escape ENV['http_proxy_pass'] || ENV['HTTP_PROXY_PASS']
@@ -742,6 +924,21 @@ class Net::HTTP::Persistent
     uri
   end
 
+  ##
+  # Returns true when proxy should by bypassed for host.
+
+  def proxy_bypass? host, port
+    host = host.downcase
+    host_port = [host, port].join ':'
+
+    @no_proxy.each do |name|
+      return true if host[-name.length, name.length] == name or
+         host_port[-name.length, name.length] == name
+    end
+
+    false
+  end
+
   ##
   # Forces reconnection of HTTP connections.
 
@@ -765,11 +962,15 @@ class Net::HTTP::Persistent
 
     finish connection
 
-    connection.start
+    start connection
   rescue Errno::ECONNREFUSED
-    raise Error, "connection refused: #{connection.address}:#{connection.port}"
+    e = Error.new "connection refused: #{connection.address}:#{connection.port}"
+    e.set_backtrace $@
+    raise e
   rescue Errno::EHOSTDOWN
-    raise Error, "host down: #{connection.address}:#{connection.port}"
+    e = Error.new "host down: #{connection.address}:#{connection.port}"
+    e.set_backtrace $@
+    raise e
   end
 
   ##
@@ -781,27 +982,14 @@ class Net::HTTP::Persistent
   #
   # +req+ must be a Net::HTTPRequest subclass (see Net::HTTP for a list).
   #
-  # If there is an error and the request is idempontent according to RFC 2616
+  # If there is an error and the request is idempotent according to RFC 2616
   # it will be retried automatically.
 
   def request uri, req = nil, &block
     retried      = false
     bad_response = false
 
-    req = Net::HTTP::Get.new uri.request_uri unless req
-
-    @headers.each do |pair|
-      req.add_field(*pair)
-    end
-
-    @override_headers.each do |name, value|
-      req[name] = value
-    end
-
-    unless req['Connection'] then
-      req.add_field 'Connection', 'keep-alive'
-      req.add_field 'Keep-Alive', @keep_alive
-    end
+    req = request_setup req || uri
 
     connection = connection_for uri
     connection_id = connection.object_id
@@ -826,23 +1014,25 @@ class Net::HTTP::Persistent
 
       bad_response = true
       retry
-    rescue IOError, EOFError, Timeout::Error,
-           Errno::ECONNABORTED, Errno::ECONNRESET, Errno::EPIPE,
-           Errno::EINVAL, OpenSSL::SSL::SSLError => e
-
-      if retried or not can_retry? req
-        due_to = "(due to #{e.message} - #{e.class})"
-        message = error_message connection
-
-        finish connection
-
-        raise Error, "too many connection resets #{due_to} #{message}"
-      end
+    rescue *RETRIED_EXCEPTIONS => e # retried on ruby 2
+      request_failed e, req, connection if
+        retried or not can_retry? req, @retried_on_ruby_2
 
       reset connection
 
       retried = true
       retry
+    rescue Errno::EINVAL, Errno::ETIMEDOUT => e # not retried on ruby 2
+      request_failed e, req, connection if retried or not can_retry? req
+
+      reset connection
+
+      retried = true
+      retry
+    rescue Exception => e
+      finish connection
+
+      raise
     ensure
       Thread.current[@timeout_key][connection_id] = Time.now
     end
@@ -852,6 +1042,51 @@ class Net::HTTP::Persistent
     response
   end
 
+  ##
+  # Raises an Error for +exception+ which resulted from attempting the request
+  # +req+ on the +connection+.
+  #
+  # Finishes the +connection+.
+
+  def request_failed exception, req, connection # :nodoc:
+    due_to = "(due to #{exception.message} - #{exception.class})"
+    message = "too many connection resets #{due_to} #{error_message connection}"
+
+    finish connection
+
+
+    raise Error, message, exception.backtrace
+  end
+
+  ##
+  # Creates a GET request if +req_or_uri+ is a URI and adds headers to the
+  # request.
+  #
+  # Returns the request.
+
+  def request_setup req_or_uri # :nodoc:
+    req = if URI === req_or_uri then
+            Net::HTTP::Get.new req_or_uri.request_uri
+          else
+            req_or_uri
+          end
+
+    @headers.each do |pair|
+      req.add_field(*pair)
+    end
+
+    @override_headers.each do |name, value|
+      req[name] = value
+    end
+
+    unless req['Connection'] then
+      req.add_field 'Connection', 'keep-alive'
+      req.add_field 'Keep-Alive', @keep_alive
+    end
+
+    req
+  end
+
   ##
   # Shuts down all connections for +thread+.
   #
