@@ -274,6 +274,7 @@ class TestNetHttpPersistent < Minitest::Test
   def test_connection_for
     @http.open_timeout = 123
     @http.read_timeout = 321
+    @http.idle_timeout = 42
     c = @http.connection_for @uri
 
     assert_kind_of @http_class, c
@@ -283,6 +284,7 @@ class TestNetHttpPersistent < Minitest::Test
 
     assert_equal 123, c.open_timeout
     assert_equal 321, c.read_timeout
+    assert_equal 42, c.keep_alive_timeout unless RUBY_1
 
     assert_includes conns[0].keys, 'example.com:80'
     assert_same c, conns[0]['example.com:80']
@@ -503,6 +505,19 @@ class TestNetHttpPersistent < Minitest::Test
     assert_same c, conns[1]['example.com:80:proxy.example:80:johndoe:muffins']
   end
 
+  def test_connection_for_proxy_unescaped
+    uri = URI.parse 'http://proxy.example'
+    uri.user = 'john%40doe'
+    uri.password = 'muf%3Afins'
+    uri.freeze
+
+    http = Net::HTTP::Persistent.new nil, uri
+    c = http.connection_for @uri
+
+    assert_includes conns[1].keys,
+                    'example.com:80:proxy.example:80:john@doe:muf:fins'
+  end
+
   def test_connection_for_proxy_host_down
     Net::HTTP.use_connect :host_down_connect
     Net::HTTP::Persistent::SSLReuse.use_connect :host_down_connect
@@ -644,6 +659,12 @@ class TestNetHttpPersistent < Minitest::Test
     assert_equal '+%3F', @http.escape(' ?')
   end
 
+  def test_unescape
+    assert_nil @http.unescape nil
+
+    assert_equal ' ?', @http.unescape('+%3F')
+  end
+
   def test_expired_eh
     c = basic_connection
     reqs[c.object_id] = 0
@@ -786,6 +807,15 @@ class TestNetHttpPersistent < Minitest::Test
     assert_equal 1, @http.ssl_generation, 'ssl_generation'
   end
 
+  def test_proxy_equals_nil
+    @http.proxy = nil
+
+    assert_equal nil, @http.proxy_uri
+
+    assert_equal 1, @http.generation, 'generation'
+    assert_equal 1, @http.ssl_generation, 'ssl_generation'
+  end
+
   def test_proxy_equals_uri
     proxy_uri = URI.parse 'http://proxy.example'
 
@@ -1337,6 +1367,7 @@ class TestNetHttpPersistent < Minitest::Test
     end
 
     assert_match %r%host down%, e.message
+    assert_match __FILE__, e.backtrace.first
   end
 
   def test_reset_io_error
@@ -1361,6 +1392,7 @@ class TestNetHttpPersistent < Minitest::Test
     end
 
     assert_match %r%connection refused%, e.message
+    assert_match __FILE__, e.backtrace.first
   end
 
   def test_retry_change_requests_equals
