@@ -1,15 +1,21 @@
 require 'rubygems'
 require 'minitest/autorun'
 require 'net/http/persistent'
-require 'openssl'
-require 'webrick'
-require 'webrick/ssl'
+have_ssl =
+  begin
+    require 'openssl'
+    require 'webrick'
+    require 'webrick/ssl'
+    true
+  rescue LoadError
+    false
+  end
 
 ##
 # This test is based on (and contains verbatim code from) the Net::HTTP tests
 # in ruby
 
-class TestNetHttpPersistentSSLReuse < MiniTest::Unit::TestCase
+class TestNetHttpPersistentSSLReuse < Minitest::Test
 
   class NullWriter
     def <<(s) end
@@ -19,9 +25,9 @@ class TestNetHttpPersistentSSLReuse < MiniTest::Unit::TestCase
   end
 
   def setup
-    @name = OpenSSL::X509::Name.parse 'CN=localhost'
+    @name = OpenSSL::X509::Name.parse 'CN=localhost/DC=localdomain'
 
-    @key = OpenSSL::PKey::RSA.new 512
+    @key = OpenSSL::PKey::RSA.new 1024
 
     @cert = OpenSSL::X509::Certificate.new
     @cert.version = 2
@@ -30,6 +36,9 @@ class TestNetHttpPersistentSSLReuse < MiniTest::Unit::TestCase
     @cert.not_after = Time.now + 300
     @cert.public_key = @key.public_key
     @cert.subject = @name
+    @cert.issuer = @name
+
+    @cert.sign @key, OpenSSL::Digest::SHA1.new
 
     @host = 'localhost'
     @port = 10082
@@ -73,12 +82,14 @@ class TestNetHttpPersistentSSLReuse < MiniTest::Unit::TestCase
   end
 
   def test_ssl_connection_reuse
+    store = OpenSSL::X509::Store.new
+    store.add_cert @cert
+
     @http = Net::HTTP::Persistent::SSLReuse.new @host, @port
+    @http.cert_store = store
+    @http.ssl_version = :SSLv3 if @http.respond_to? :ssl_version=
     @http.use_ssl = true
     @http.verify_mode = OpenSSL::SSL::VERIFY_PEER
-    @http.verify_callback = proc do |_, store_ctx|
-      store_ctx.current_cert.to_der == @cert.to_der
-    end
 
     @http.start
     @http.get '/'
@@ -97,5 +108,5 @@ class TestNetHttpPersistentSSLReuse < MiniTest::Unit::TestCase
     assert ssl_socket.session_reused?
   end
 
-end
+end if have_ssl
 
